import typing as t

from sqlalchemy import Select, select

from .base import _select_stmt_fields

_T = t.TypeVar("_T")


def find(
    cls: t.Union[t.Any, t.Tuple],
    offset: t.Optional[int] = None,
    rows: t.Optional[int] = None,
    **kwargs
) -> Select:
    """Generate select statement with filters.

    Args:
        cls (t.Union[t.Any, t.Tuple]): table class
        offset (t.Optional[int], optional): number of rows to skip. Defaults to None.
        rows (t.Optional[int], optional): number of rows to return. Defaults to None.
        **kwargs: search filters

    Returns:
        Select: _description_
    """
    select_stmt = _select_stmt_fields(fields=cls, offset=offset, rows=rows)
    stmt = select_stmt.filter_by(**kwargs)
    return stmt


def all(cls) -> Select:
    """Generate select to return all items.

    Returns:
        Select: select statement
    """
    stmt = _select_stmt_fields(fields=cls)
    return stmt


def exists(cls, **kwargs) -> Select:
    """Generate select statement to check if item exists.

    Returns:
        Select: select statement
    """
    stmt = select(cls).filter_by(**kwargs)
    return stmt



import typing as t
from functools import wraps

from crudal.types import CRUDALType

T = t.TypeVar("T", bound=CRUDALType)
_RT = t.TypeVar("_RT")  # return type
P = t.ParamSpec("P")


def with_session_sync(f: t.Callable[P, _RT]) -> t.Callable[P, _RT]:
    """Decorator to handle session."""

    @wraps(f)
    def wrapper(*args: P.args, **kwargs: P.kwargs):
        ref = args[0]
        if len(args) >= 2:
            session = args[1]
        else:
            session = None

        # if session is passed as argument
        # use it
        if session is not None:
            return f(ref, session, **kwargs)

        # if session is not passed as argument
        # and class has __session__ attribute
        elif session is None and ref.__session__ is not None:
            with ref.__session__() as session:
                return f(ref, session, **kwargs)
        else:
            raise ValueError("Neither function session or class session exists")

    return wrapper


def with_session_async(f: t.Callable[P, _RT]) -> t.Callable[P, _RT]:
    """Decorator to handle session."""

    @wraps(f)
    async def wrapper(*args: P.args, **kwargs: P.kwargs):
        ref = args[0]
        if len(args) >= 2:
            session = args[1]
        else:
            session = None

        # if session is passed as argument
        if session is not None:
            return await f(ref, session, **kwargs)

        # if session is not passed as argument
        # and class has __session__ attribute
        elif session is None and ref.__session__ is not None:
            async with ref.__session__() as session:
                return await f(ref, session, **kwargs)
        else:
            raise ValueError("Neither function session or class session exists")

    return wrapper



import typing as t

from sqlalchemy import select


def _select_stmt_fields(
    fields, rows: t.Optional[int] = None, offset: t.Optional[int] = None
):
    """Generate select statement with fields.

    Args:
        fields: select fields
        rows (t.Optional[int], optional): number of rows to return. Defaults to None.
        offset (t.Optional[int], optional): number of rows to skip. Defaults to None.

    Returns:
        Select: select statement
    """
    # generate docstring
    if isinstance(t, tuple) or isinstance(t, list):
        stmt = select(*fields)
    else:
        stmt = select(fields)

    if offset is not None:
        stmt = stmt.offset(offset)

    if rows is not None:
        stmt = stmt.limit(rows)

    return stmt



from sqlalchemy import Update, update


def update_(cls, values: dict, **filters) -> Update:
    """Generate update statement

    Args:
        values (dict): values to update

    Returns:
        Update: update statement
    """
    stmt = update(cls).filter_by(**filters).values(**values)
    return stmt



import typing as t

from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Session

_T = t.TypeVar("_T")


class CRUDALType(t.Protocol):
    @classmethod
    def _get_primary_key(cls) -> str:
        """Table primary key"""
        raise NotImplementedError()

    @classmethod
    def find(
        cls: t.Type[_T],
        session: t.Optional[AsyncSession] = None,
        /,
        *,
        rows: t.Optional[int] = None,
        offset: int = 0,
        **filters,
    ) -> t.Sequence[_T]:
        """Find items"""
        raise NotImplementedError()

    @classmethod
    def find_by_pk(cls: t.Type[_T], pk: t.Any, session: Session) -> t.Optional[_T]:
        raise NotImplementedError()

    @classmethod
    def exists(cls, session: Session, **filters) -> bool:
        raise NotImplementedError()

    @classmethod
    def all(cls: t.Type[_T], session: Session) -> t.Sequence[_T]:
        raise NotImplementedError()

    @classmethod
    def delete(cls, session: Session, **filters) -> bool:
        raise NotImplementedError()

    @classmethod
    def add_many(cls: t.Type[_T], session: Session, items: t.List[_T]) -> None:
        raise NotImplementedError()

    def add(self, session: Session, commit: bool = False) -> None:
        raise NotImplementedError()


class CRUDALTypeAsync(t.Protocol):
    @classmethod
    def _session(cls):
        raise NotImplementedError()

    @classmethod
    def _get_primary_key(cls) -> str:
        """Table primary key"""
        raise NotImplementedError()

    @classmethod
    async def find(
        cls: t.Type[_T],
        session: t.Optional[AsyncSession] = None,
        /,
        *,
        rows: t.Optional[int] = None,
        offset: int = 0,
        **filters,
    ) -> t.Sequence[_T]:
        """Find items"""
        raise NotImplementedError()

    @classmethod
    async def find_by_pk(
        cls: t.Type[_T], pk: t.Any, session: AsyncSession
    ) -> t.Optional[_T]:
        raise NotImplementedError()

    @classmethod
    async def exists(cls, session: AsyncSession, **filters) -> bool:
        raise NotImplementedError()

    @classmethod
    async def all(cls: t.Type[_T], session: AsyncSession) -> t.Sequence[_T]:
        raise NotImplementedError()

    @classmethod
    async def delete(cls, session: AsyncSession, **filters) -> bool:
        raise NotImplementedError()

    @classmethod
    async def add_many(
        cls: t.Type[_T], session: AsyncSession, items: t.List[_T]
    ) -> None:
        raise NotImplementedError()

    async def add(self, session: AsyncSession, commit: bool = False) -> None:
        raise NotImplementedError()



from sqlalchemy import Delete, delete


def delete_(cls, **kwargs) -> Delete:
    """Generate delete statement

    Returns:
        Delete: delete statement
    """
    stmt = delete(cls).filter_by(**kwargs)
    return stmt



import typing as t

from sqlalchemy import Result, ScalarResult
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.inspection import inspect
from sqlalchemy.orm import DeclarativeBase

from crudal import operations
from crudal.types import CRUDALTypeAsync
from crudal.utils import with_session_async

_T = t.TypeVar("_T", bound=CRUDALTypeAsync)


async def _crud_stmt_scalars(stmt, session: AsyncSession) -> ScalarResult:
    return await session.scalars(stmt)


async def _crud_stmt_execute(stmt, session: AsyncSession) -> Result:
    return await session.execute(stmt)


class DeclarativeCrudBaseAsync(DeclarativeBase):
    __mapper_args__ = {"eager_defaults": True}
    __session__ = None

    @classmethod
    def _get_primary_key(cls) -> str:
        """Return PK column name"""
        return inspect(cls).primary_key[0].name

    @classmethod
    @with_session_async
    async def find(
        cls: t.Type[_T],
        session: AsyncSession,
        /,
        *,
        rows: t.Optional[int] = None,
        offset: int = 0,
        **filters,
    ) -> t.Sequence[_T]:
        """Find items in table.

        Example:
        ```
        # find all users with name Andrew
        await User.find(session, name="Andrew")
        ```

        Args:
            session (AsyncSession): SQLAlchemy session
            rows(int, optional): Number of rows to return. Defaults to None.
            offset(int, optional): Number of rows to skip. Defaults to 0.
            **filters: search filters

        Returns:
            t.Sequence[_T]: Found items
        """
        stmt = operations.find(cls, offset=offset, rows=rows, **filters)
        result = await _crud_stmt_scalars(stmt, session=session)
        return result.all()

    @classmethod
    @with_session_async
    async def find_by_pk(
        cls: t.Type[_T], session: AsyncSession, /, *, pk: t.Any
    ) -> t.Optional[_T]:
        """Find row by its primary key

        Example:
        ```
        # find user with id 1
        await User.find_by_pk(session, pk=1)
        ```

        Args:
            session (AsyncSession): SQLAlchemy session
            pk (t.Any): primary key value

        Raises:
            ValueError: Multiple rows found by one primary key

        Returns:
            t.Optional[_T]: found item.
                If None - no items with such primary keys exists
        """
        pk_col = cls._get_primary_key()
        result = await cls.find(session, **{pk_col: pk})
        if len(result) == 1:
            return result[0]
        elif len(result) > 1:
            raise ValueError(
                f"Found more than one items ({len(result)}) with {pk_col}={pk}"
            )
        else:
            return None

    @classmethod
    @with_session_async
    async def exists(cls, session: AsyncSession, /, **filters) -> bool:
        """Check if items exists in table

        Example:
        ```
        # check if user with name Andrew exists
        await User.exists(session, name="Andrew")
        ```

        Args:
            session (AsyncSession): SQLAlchemy session
            **filters: search filters

        Returns:
            bool: True if exists, False if not
        """
        stmt = operations.find(cls, **filters)
        result = await _crud_stmt_scalars(stmt, session=session)
        return True if result.first() is not None else False

    @classmethod
    @with_session_async
    async def all(cls: t.Type[_T], session: AsyncSession, /) -> t.Sequence[_T]:
        """Get all table items

        Example:
        ```
        # get all users
        await User.all(session)
        ```

        Args:
            session (AsyncSession): SQLAlchemy session

        Returns:
            t.Sequence[_T]: all table items

        """
        stmt = operations.find(cls)
        result = await _crud_stmt_scalars(stmt, session=session)
        return result.all()

    @classmethod
    @with_session_async
    async def delete(
        cls, session: AsyncSession, /, commit: bool = False, **filters
    ) -> bool:
        """Delete items from table

        Example:
        ```
        # delete all users with name Andrew
        await User.delete(session, name="Andrew")
        ```

        Args:
            session (AsyncSession): SQLAlchemy session
            commit (bool, optional): commit after deleting. Defaults to False.

        Returns:
            bool: _description_
        """
        exists = await cls.exists(session, **filters)
        if not exists:
            return False

        delete_stmt = operations.delete_(cls=cls, **filters)
        await session.execute(delete_stmt)

        if commit:
            await session.commit()

        return True

    @classmethod
    @with_session_async
    async def add_many(
        cls: t.Type[_T],
        session: AsyncSession,
        /,
        *,
        items: t.List[_T],
        commit: bool = False,
    ) -> None:
        """Add many new items to table

        Example:
        ```
        users = [User(name="Andrew"), User(name="Bob")]
        await User.add_many(session, items=users)
        ```

        Args:
            session (AsyncSession): SQLAlchemy session
            items (t.List[_T]): list of items to add
            commit (bool, optional): commit after adding. Defaults to False.
        """
        session.add_all(items)
        if commit:
            await session.commit()

    @with_session_async
    async def add(self: _T, session: AsyncSession, /, *, commit: bool = False) -> _T:
        """Add one item to table.

        Example:
        ```
        user = User(name="Andrew")
        await user.add(session)
        ```

        Args:
            session (AsyncSession): SQLAlchemy async session
            commit (bool, optional): Commit or not. Defaults to False.
        """
        session.add(self)
        if commit:
            await session.commit()
            await session.refresh(self)

        return self

    @classmethod
    async def update(
        cls: t.Type[_T], session: AsyncSession, /, *, values: dict, **filters
    ):
        """Update table items values

        Example:
        ```
        # update all users with name Andrew to name Bob
        await User.update(session, values={"name": "Bob"}, name="Andrew")
        ```

        Args:
            session (AsyncSession): SQLAlchemy session
            values (dict): values to update
            **filters: search filters
        """
        stmt = operations.update_(cls, values=values, **filters)
        return await _crud_stmt_execute(stmt=stmt, session=session)
